# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
import numpy as np
from PySide6 import QtWidgets, QtGui
from matplotlib import pyplot as plt
from typing import Union
from VeraGrid.Gui.table_view_header_wrap import HeaderViewWithWordWrap
import VeraGrid.Gui.gui_functions as gf
from VeraGrid.Gui.messages import error_msg, warning_msg, yes_no_question
from VeraGrid.Gui.Main.SubClasses.simulations import SimulationsMain
from VeraGrid.Gui.results_model import ResultsModel
from VeraGrid.Gui.general_dialogues import fill_tree_from_logs
from VeraGrid.Gui.rms_plot_variables_dialog import RmsPlotDialog
import VeraGridEngine.Utils.Filtering as flt
from VeraGridEngine.basic_structures import Logger
from VeraGridEngine.enumerations import ResultTypes


class ResultsMain(SimulationsMain):
    """
    Diagrams Main
    """

    def __init__(self, parent=None):
        """

        @param parent:
        """

        # create main window
        SimulationsMain.__init__(self, parent)

        self.results_mdl: Union[None, ResultsModel] = None

        self.current_results_logger: Union[None, Logger] = None

        # --------------------------------------------------------------------------------------------------------------
        self.ui.actionSet_OPF_generation_to_profiles.triggered.connect(self.copy_opf_to_profiles)

        # Buttons
        self.ui.saveResultsButton.clicked.connect(self.save_results_df)
        self.ui.copy_results_pushButton.clicked.connect(self.copy_results_data)
        self.ui.copy_numpy_button.clicked.connect(self.copy_results_data_as_numpy)
        self.ui.plot_data_pushButton.clicked.connect(self.plot_results)
        self.ui.plot_dyn_data_pushButton.clicked.connect(self.plot_dyn_results)
        self.ui.search_results_Button.clicked.connect(self.search_in_results)
        self.ui.deleteDriverButton.clicked.connect(self.delete_results_driver)
        self.ui.saveResultsLogsButton.clicked.connect(self.save_results_logs)

        # tree-click
        self.ui.results_treeView.clicked.connect(self.results_tree_view_click)

        # line edit enter
        self.ui.search_results_lineEdit.returnPressed.connect(self.search_in_results)

        # wrap headers
        self.ui.resultsTableView.setHorizontalHeader(HeaderViewWithWordWrap(self.ui.resultsTableView))

    def results_tree_view_click(self, index: QtGui.QStandardItem):
        """
        Display the simulation results on the result's table
        :param index: Clicked Tree index
        """
        tree_mdl = self.ui.results_treeView.model()
        item = tree_mdl.itemFromIndex(index)
        path = gf.get_tree_item_path(item)

        if len(path) > 0:
            study_name = path[0]
            driver = self.session.get_driver_by_name(study_name=study_name)

            if driver is None:
                # set the logs
                self.current_results_logger = None
                self.ui.resultsLogsTreeView.setModel(None)
                return

            # set the logs
            self.current_results_logger = driver.logger
            logs_mdl = fill_tree_from_logs(driver.logger)
            self.ui.resultsLogsTreeView.setModel(logs_mdl)
            self.ui.resultsLogsTreeView.expandAll()

            # set the report
            self.ui.resultsReportTextEdit.setText(driver.results.report_text)

            if len(path) > 1:

                if len(path) == 2:
                    result_name = path[1]
                elif len(path) == 3:
                    result_name = path[2]
                else:
                    raise Exception('Path len ' + str(len(path)) + ' not supported')

                study_results = self.available_results_dict.get(study_name, None)

                if study_results is not None:

                    study_type: ResultTypes = study_results.get(result_name, None)

                    if study_type is not None:

                        self.results_mdl = self.session.get_results_model_by_name(study_name=study_name,
                                                                                  study_type=study_type)

                        if self.results_mdl is not None:

                            # pass the matching list of devices to the ResultsModel and ResultsTable for filtering
                            self.results_mdl.table.set_col_devices(
                                devices_list=self.circuit.get_elements_by_type(self.results_mdl.table.cols_device_type)
                            )
                            self.results_mdl.table.set_idx_devices(
                                devices_list=self.circuit.get_elements_by_type(self.results_mdl.table.idx_device_type)
                            )

                            if self.ui.results_traspose_checkBox.isChecked():
                                self.results_mdl.transpose()

                            if self.ui.results_as_abs_checkBox.isChecked():
                                self.results_mdl.convert_to_abs()

                            if self.ui.results_as_cdf_checkBox.isChecked():
                                self.results_mdl.convert_to_cdf()

                            # set the table model
                            self.ui.resultsTableView.setModel(self.results_mdl)
                            self.ui.units_label.setText(self.results_mdl.units)

                        else:
                            self.ui.resultsTableView.setModel(None)
                            self.ui.units_label.setText("")

                    else:
                        self.ui.resultsTableView.setModel(None)
                        self.ui.units_label.setText("")

                else:
                    self.ui.resultsTableView.setModel(None)
                    self.ui.units_label.setText("")

            else:
                pass

        else:
            # set the logs
            self.current_results_logger = None
            self.ui.resultsLogsTreeView.setModel(None)

    def plot_results(self):
        """
        Plot the results
        """
        mdl: ResultsModel = self.ui.resultsTableView.model()

        if mdl is not None:

            plt.rcParams["date.autoformatter.minute"] = "%Y-%m-%d %H:%M:%S"

            # get the selected element
            obj_idx = self.ui.resultsTableView.selectedIndexes()
            n_cols = mdl.table.c

            if n_cols > 50:
                ok = yes_no_question(text=f"There are {n_cols} columns, the plot might take a lot to render.\n"
                                          "Are you ok with potentially waiting a lot?", title="Plot")
            else:
                ok = True

            if ok:
                # create figure to plot
                fig = plt.figure(figsize=(12, 8))
                ax = fig.add_subplot(111)

                if len(obj_idx):

                    # get the unique columns in the selected cells
                    cols = np.zeros(len(obj_idx), dtype=int)
                    rows = np.zeros(len(obj_idx), dtype=int)

                    for i in range(len(obj_idx)):
                        cols[i] = obj_idx[i].column()
                        rows[i] = obj_idx[i].row()

                    cols = np.unique(cols)
                    rows = np.unique(rows)

                else:
                    # plot all
                    cols = None
                    rows = None

                # none selected, plot all
                mdl.plot(
                    ax=ax,
                    selected_col_idx=cols,
                    selected_rows=rows,
                    stacked=self.ui.stacked_plot_checkBox.isChecked()
                )

                plt.show()
            else:
                pass

    def plot_dyn_results(self):
        """
        Show the dynamic results plot wizard
        """

        drv, results = self.session.rms_dynamic_simulation

        if results:

            dlg = RmsPlotDialog(grid=self.circuit, results=results)
            dlg.exec()

        else:
            self.show_warning_toast("No RMS simulation")

    def save_results_df(self):
        """
        Save the data displayed at the results as excel
        """
        mdl: ResultsModel = self.ui.resultsTableView.model()

        if mdl is not None:
            file, filter_ = QtWidgets.QFileDialog.getSaveFileName(self, "Export results", '',
                                                                  filter="CSV (*.csv);;Excel files (*.xlsx)")

            if file != '':
                if 'xlsx' in filter_:
                    f = file
                    if not f.endswith('.xlsx'):
                        f += '.xlsx'
                    mdl.save_to_excel(f)
                    print('Saved!')
                if 'csv' in filter_:
                    f = file
                    if not f.endswith('.csv'):
                        f += '.csv'
                    mdl.save_to_csv(f)
                    print('Saved!')
                else:
                    error_msg(file[0] + ' is not valid :(')
        else:
            warning_msg('There is no profile displayed, please display one', 'Copy profile to clipboard')

    def copy_results_data(self):
        """
        Copy the current displayed profiles to the clipboard
        """
        mdl = self.ui.resultsTableView.model()
        if mdl is not None:
            mdl.copy_to_clipboard()
            self.show_info_toast('Copied!')
        else:
            warning_msg('There is no profile displayed, please display one', 'Copy profile to clipboard')

    def copy_results_data_as_numpy(self):
        """
        Copy the current displayed profiles to the clipboard
        """
        mdl = self.ui.resultsTableView.model()
        if mdl is not None:
            mdl.copy_numpy_to_clipboard()
            self.show_info_toast('Copied!')
        else:
            warning_msg('There is no profile displayed, please display one', 'Copy profile to clipboard')

    def search_in_results(self):
        """
        Search in the results model
        """

        if self.results_mdl is not None:

            txt = self.ui.search_results_lineEdit.text().strip()

            filter_ = flt.FilterResultsTable(self.results_mdl.table)

            try:
                filter_.parse(expression=txt)
                filtered_table = filter_.apply()
            except ValueError as e:
                error_msg(str(e), "Fiter parse")
                return None
            except Exception as e:
                error_msg(str(e), "Fiter parse")
                return None

            self.results_mdl = ResultsModel(filtered_table)
            self.ui.resultsTableView.setModel(self.results_mdl)
        else:
            return None

    def delete_results_driver(self):
        """
        Delete the driver
        :return:
        """
        idx = self.ui.results_treeView.selectedIndexes()
        if len(idx) > 0:
            tree_mdl = self.ui.results_treeView.model()
            item = tree_mdl.itemFromIndex(idx[0])
            path = gf.get_tree_item_path(item)

            if len(path) > 0:
                study_name = path[0]
                study_type = self.available_results_dict[study_name]

                quit_msg = "Do you want to delete_with_dialogue the results driver " + study_name + "?"
                reply = QtWidgets.QMessageBox.question(self, 'Message',
                                                       quit_msg,
                                                       QtWidgets.QMessageBox.StandardButton.Yes,
                                                       QtWidgets.QMessageBox.StandardButton.No)

                if reply == QtWidgets.QMessageBox.StandardButton.Yes.value:
                    self.session.delete_driver_by_name(study_name)
                    self.update_available_results()

    def copy_opf_to_profiles(self):
        """
        Copy the results from the OPF snapshot and time series to the database
        """

        # copy the snapshot if that exits
        _, results = self.session.optimal_power_flow
        if results is not None:

            ok = yes_no_question('Are you sure that you want to overwrite '
                                 'the generation, batteries and load snapshot values '
                                 'with the OPF results?',
                                 title="Overwrite profiles with OPF results")

            if ok:
                self.circuit.set_opf_snapshot_results(results)
                self.show_info_toast("P snapshot set from the OPF results")

        else:
            self.show_warning_toast('The OPF time series has no results :(')

        # copy the time series if that exists --------------------------------------------------------------------------
        _, results = self.session.optimal_power_flow_ts
        if results is not None:

            ok = yes_no_question('Are you sure that you want to overwrite '
                                 'the generation, batteries and load profiles '
                                 'with the OPF time series results?',
                                 title="Overwrite profiles with OPF results")

            if ok:
                self.circuit.set_opf_ts_results(results)
                self.show_info_toast("P profiles set from the OPF results")

        else:
            self.show_warning_toast('The OPF time series has no results :(')

    def save_results_logs(self):
        """
        Save the results' logs
        """
        file, filter_ = QtWidgets.QFileDialog.getSaveFileName(self, "Export logs", '',
                                                              filter="CSV (*.csv);;Excel files (*.xlsx)", )

        if file != '':
            if 'xlsx' in filter_:
                f = file
                if not f.endswith('.xlsx'):
                    f += '.xlsx'
                self.current_results_logger.to_xlsx(f)

            if 'csv' in filter_:
                f = file
                if not f.endswith('.csv'):
                    f += '.csv'
                self.current_results_logger.to_csv(f)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.  
# SPDX-License-Identifier: MPL-2.0

from __future__ import annotations
from typing import TYPE_CHECKING, Union
from PySide6.QtGui import QIcon, QPixmap
from PySide6.QtWidgets import QMenu
from VeraGrid.Gui.Diagrams.SchematicWidget.terminal_item import BarTerminalItem, RoundTerminalItem
from VeraGridEngine.Devices.Branches.series_reactance import SeriesReactance
from VeraGrid.Gui.Diagrams.SchematicWidget.Branches.line_graphics_template import LineGraphicTemplateItem

if TYPE_CHECKING:  # Only imports the below statements during type checking
    from VeraGrid.Gui.Diagrams.SchematicWidget.schematic_widget import SchematicWidget


class SeriesReactanceGraphicItem(LineGraphicTemplateItem):

    def __init__(self,
                 from_port: Union[BarTerminalItem, RoundTerminalItem],
                 to_port: Union[BarTerminalItem, RoundTerminalItem],
                 editor: SchematicWidget,
                 width=5,
                 api_object: SeriesReactance = None,
                 draw_labels: bool = True):
        """

        :param from_port:
        :param to_port:
        :param editor:
        :param width:
        :param api_object:
        """
        LineGraphicTemplateItem.__init__(self=self,
                                         from_port=from_port,
                                         to_port=to_port,
                                         editor=editor,
                                         width=width,
                                         api_object=api_object,
                                         draw_labels=draw_labels)

    @property
    def api_object(self) -> SeriesReactance:
        return self._api_object

    def contextMenuEvent(self, event):
        """
        Show context menu
        @param event:
        @return:
        """
        if self.api_object is not None:
            menu = QMenu()

            pe = menu.addAction('Enable/Disable')
            pe_icon = QIcon()
            if self.api_object.active:
                pe_icon.addPixmap(QPixmap(":/Icons/icons/uncheck_all.png"))
            else:
                pe_icon.addPixmap(QPixmap(":/Icons/icons/check_all.png"))
            pe.setIcon(pe_icon)
            pe.triggered.connect(self.enable_disable_toggle)

            rabf = menu.addAction('Change bus')
            move_bus_icon = QIcon()
            move_bus_icon.addPixmap(QPixmap(":/Icons/icons/move_bus.png"))
            rabf.setIcon(move_bus_icon)
            rabf.triggered.connect(self.change_bus)

            menu.addSeparator()

            ra2 = menu.addAction('Delete')
            del_icon = QIcon()
            del_icon.addPixmap(QPixmap(":/Icons/icons/delete3.png"))
            ra2.setIcon(del_icon)
            ra2.triggered.connect(self.delete)

            menu.addSeparator()

            ra6 = menu.addAction('Plot profiles')
            plot_icon = QIcon()
            plot_icon.addPixmap(QPixmap(":/Icons/icons/plot.png"))
            ra6.setIcon(plot_icon)
            ra6.triggered.connect(self.plot_profiles)

            ra4 = menu.addAction('Assign rate to profile')
            ra4_icon = QIcon()
            ra4_icon.addPixmap(QPixmap(":/Icons/icons/assign_to_profile.png"))
            ra4.setIcon(ra4_icon)
            ra4.triggered.connect(self.assign_rate_to_profile)

            ra5 = menu.addAction('Assign active state to profile')
            ra5_icon = QIcon()
            ra5_icon.addPixmap(QPixmap(":/Icons/icons/assign_to_profile.png"))
            ra5.setIcon(ra5_icon)
            ra5.triggered.connect(self.assign_status_to_profile)


            menu.exec_(event.screenPos())
        else:
            pass



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations
import numpy as np
import pandas as pd
from typing import Union, Any
from PySide6 import QtCore, QtWidgets
from VeraGrid.Gui.wrappable_table_model import WrappableTableModel
from VeraGridEngine.enumerations import ResultTypes


def fast_data_to_numpy_text(data: np.ndarray):
    """
    Convert numpy to text
    :param data: numpy array
    :return:
    """
    if len(data.shape) == 1:
        txt = '[' + ', '.join(['{0:.6f}'.format(x) for x in data]) + ']'

    elif len(data.shape) == 2:

        if data.shape[1] > 1:
            # header first
            txt = '['

            # data
            for t in range(data.shape[0]):
                txt += '[' + ', '.join(['{0:.6f}'.format(x) for x in data[t, :]]) + '],\n'

            txt += ']'
        else:
            txt = '[' + ', '.join(['{0:.6f}'.format(x) for x in data[:, 0]]) + ']'
    else:
        txt = '[]'

    return txt


class PandasModel(WrappableTableModel):
    """
    Class to populate a Qt table view with a pandas data frame
    """

    def __init__(self,
                 data: pd.DataFrame,
                 parent: QtWidgets.QTableView = None,
                 editable=False,
                 editable_min_idx=-1,
                 decimals=6):
        """

        :param data:
        :param parent:
        :param editable:
        :param editable_min_idx:
        :param decimals:
        """
        WrappableTableModel.__init__(self, parent)
        self.data_c = data.values
        self.cols_c = data.columns
        self.index_c = data.index.values
        self.editable = editable
        self.editable_min_idx = editable_min_idx
        self.r, self.c = self.data_c.shape
        self.isDate = False
        if self.r > 0 and self.c > 0:
            if isinstance(self.index_c[0], np.datetime64):
                self.index_c = pd.to_datetime(self.index_c)
                self.isDate = True

        self.format_string = '.' + str(decimals) + 'f'

        self.formatter = lambda x: "%.2f" % x

    def flags(self, index: QtCore.QModelIndex):
        """

        :param index:
        :return:
        """
        if self.editable and index.column() > self.editable_min_idx:
            return (QtCore.Qt.ItemFlag.ItemIsEditable |
                    QtCore.Qt.ItemFlag.ItemIsEnabled |
                    QtCore.Qt.ItemFlag.ItemIsSelectable)
        else:
            return QtCore.Qt.ItemFlag.ItemIsEnabled

    def rowCount(self, parent: Union[QtCore.QModelIndex, QtCore.QPersistentModelIndex] = ...) -> int:
        """

        :param parent:
        :return:
        """
        return self.r

    def columnCount(self, parent: Union[QtCore.QModelIndex, QtCore.QPersistentModelIndex] = ...) -> int:
        """

        :param parent:
        :return:
        """
        return self.c

    def data(self, index: QtCore.QModelIndex, role: int = QtCore.Qt.ItemDataRole.DisplayRole) -> Any:
        """

        :param index:
        :param role:
        :return:
        """
        if index.isValid() and role == QtCore.Qt.ItemDataRole.DisplayRole:

            val = self.data_c[index.row(), index.column()]
            if isinstance(val, str):
                return val
            elif isinstance(val, complex):
                if val.real != 0 or val.imag != 0:
                    return val.__format__(self.format_string)
                else:
                    return '0'
            else:
                if val != 0:
                    return val.__format__(self.format_string)
                else:
                    return '0'
        return None

    def setData(self, index, value, role=QtCore.Qt.ItemDataRole.DisplayRole):
        """

        :param index:
        :param value:
        :param role:
        :return:
        """
        self.data_c[index.row(), index.column()] = value
        return None

    def headerData(self,
                   section: int,
                   orientation: QtCore.Qt.Orientation,
                   role=QtCore.Qt.ItemDataRole.DisplayRole):
        """

        :param section:
        :param orientation:
        :param role:
        :return:
        """
        if self._hide_headers_mode is True:
            return None

        if role == QtCore.Qt.ItemDataRole.DisplayRole:
            if orientation == QtCore.Qt.Orientation.Horizontal:
                return self.cols_c[section]
            elif orientation == QtCore.Qt.Orientation.Vertical:
                if self.index_c is None:
                    return section
                else:
                    if self.isDate:
                        return self.index_c[section].strftime('%Y/%m/%d  %H:%M.%S')
                    else:
                        return str(self.index_c[section])
        return None

    def copy_to_column(self, row, col):
        """
        Copies one value to all the column
        @param row: Row of the value
        @param col: Column of the value
        @return: Nothing
        """
        self.data_c[:, col] = self.data_c[row, col]

    def is_complex(self) -> bool:
        """
        id the data type complex?
        :return: True / False
        """
        return self.data_c.dtype == complex

    def is_2d(self) -> bool:
        """
        actually check if the array is 1D or 2D
        :return: true if it is really a 2D data set
        """

        is_2d_ = len(self.data_c.shape) == 2
        if is_2d_:
            if self.data_c.shape[1] <= 1:
                is_2d_ = False
        return is_2d_

    def get_data(self, mode=None):
        """

        :param mode: 'real', 'imag', 'abs'
        :return: index, columns, data
        """

        n = len(self.cols_c)

        if n > 0:
            # gather values
            if isinstance(self.cols_c, pd.Index):
                names = self.cols_c.values

                if len(names) > 0:
                    if isinstance(names[0], ResultTypes):
                        names = [val.name for val in names]

            # elif isinstance(self.cols_c, ResultTypes):
            #     names = [val.value for val in self.cols_c]

            else:
                names = [val.name for val in self.cols_c]

            if self.is_complex():

                if mode == 'real':
                    values = self.data_c.real
                elif mode == 'imag':
                    values = self.data_c.imag
                elif mode == 'abs':
                    values = np.abs(self.data_c)
                else:
                    values = self.data_c.astype(object)

            else:
                values = self.data_c

            return self.index_c, names, values
        else:
            # there are no elements
            return list(), list(), list()

    def get_df(self, mode=None):
        """
        Get the data as pandas DataFrame
        :return: DataFrame
        """
        index, columns, data = self.get_data(mode=mode)
        return pd.DataFrame(data=data, index=index, columns=columns)

    def save_to_excel(self, file_name, mode):
        """

        :param file_name:
        :param mode:
        :return:
        """
        index, columns, data = self.get_data(mode=mode)

        df = pd.DataFrame(data=data, index=index, columns=columns)
        df.to_excel(file_name)

    def copy_to_clipboard(self, mode=None):
        """
        Copy profiles to clipboard
        :param mode: numpy or other -> separated by tabs
        :return:
        """
        n = len(self.cols_c)

        if n > 0:

            if mode == 'numpy':
                txt = fast_data_to_numpy_text(self.data_c)
            else:
                index, columns, data = self.get_data(mode=mode)

                data = data.astype(str)

                # header first
                txt = '\t' + '\t'.join(columns) + '\n'

                # data
                for t, index_value in enumerate(index):
                    txt += str(index_value) + '\t' + '\t'.join(data[t, :]) + '\n'

            # copy to clipboard
            cb = QtWidgets.QApplication.clipboard()
            cb.clear()
            cb.setText(txt)



# MIT License
#
# Copyright (c) 2021-2022 Yunosuke Ohsugi
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""This script is inspired by darkdetect(https://github.com/albertosottile/darkdetect)."""
from __future__ import annotations

import ctypes
from ctypes import c_void_p

try:
    # macOS Big Sur+ use "a built-in dynamic linker cache of all system-provided libraries"
    _objc = ctypes.cdll.LoadLibrary("libobjc.dylib")
except OSError:
    import ctypes.util

    # revert to full path for older OS versions and hardened programs
    _objc = ctypes.cdll.LoadLibrary(ctypes.util.find_library("objc"))  # type: ignore

_objc.objc_getClass.restype = c_void_p
_objc.sel_registerName.restype = c_void_p

msg_prototype = ctypes.CFUNCTYPE(c_void_p, c_void_p, c_void_p, c_void_p)
msg = msg_prototype(("objc_msgSend", _objc), ((1, "", None), (1, "", None), (1, "", None)))

_MAC_ACCENT_COLORS = {
    None: None,
    "-1": "graphite",
    "0": "red",
    "1": "orange",
    "2": "yellow",
    "3": "green",
    "4": "blue",
    "5": "purple",
    "6": "pink",
}


def _utf8(s: str | bytes):
    return s.encode("utf8") if not isinstance(s, bytes) else s


def _n(name: str):
    return _objc.sel_registerName(_utf8(name))


def _c(classname: str):
    return _objc.objc_getClass(_utf8(classname))


def get_mac_accent() -> str | None:
    """Get system accent color on Mac."""
    pool = msg(_c("NSAutoreleasePool"), _n("alloc"))
    pool = msg(pool, _n("init"))

    std_user_def = msg(_c("NSUserDefaults"), _n("standardUserDefaults"))

    key = msg(_c("NSString"), _n("stringWithUTF8String:"), _utf8("AppleAccentColor"))
    accent_id = msg(std_user_def, _n("stringForKey:"), c_void_p(key))
    accent_id = msg(accent_id, _n("UTF8String"))

    accent_id = None if accent_id is None else ctypes.string_at(accent_id).decode()

    msg(pool, _n("release"))
    return _MAC_ACCENT_COLORS.get(accent_id)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.  
# SPDX-License-Identifier: MPL-2.0

from __future__ import annotations
from typing import TYPE_CHECKING
from PySide6.QtGui import QPen, QIcon, QPixmap
from PySide6.QtWidgets import QMenu, QGraphicsTextItem
from VeraGridEngine.Devices.Injections.static_generator import StaticGenerator
from VeraGrid.Gui.Diagrams.generic_graphics import ACTIVE, DEACTIVATED, OTHER, Square, Condenser
from VeraGrid.Gui.Diagrams.MapWidget.Injections.map_injections_template_graphics import MapInjectionTemplateGraphicItem
from VeraGrid.Gui.messages import yes_no_question

if TYPE_CHECKING:  # Only imports the below statements during type checking
    from VeraGrid.Gui.Diagrams.MapWidget.grid_map_widget import GridMapWidget


class MapStaticGeneratorGraphicItem(MapInjectionTemplateGraphicItem):

    def __init__(self, api_object: StaticGenerator,
                 editor: GridMapWidget,
                 lat: float,
                 lon: float,
                 size: float = 0.8,
                 draw_labels: bool = True):
        """

        :param api_object:
        :param editor:
        """
        MapInjectionTemplateGraphicItem.__init__(self,
                                                 api_object=api_object,
                                                 editor=editor,
                                                 lat=lat,
                                                 lon=lon,
                                                 size=size,
                                                 draw_labels=draw_labels)
    @property
    def api_object(self) -> StaticGenerator:
        return self._api_object



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'solar_power_wizard_gui.ui'
##
## Created by: Qt User Interface Compiler version 6.9.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QAction, QBrush, QColor, QConicalGradient,
    QCursor, QFont, QFontDatabase, QGradient,
    QIcon, QImage, QKeySequence, QLinearGradient,
    QPainter, QPalette, QPixmap, QRadialGradient,
    QTransform)
from PySide6.QtWidgets import (QApplication, QDialog, QDoubleSpinBox, QFrame,
    QGridLayout, QHeaderView, QLabel, QPushButton,
    QSizePolicy, QSpacerItem, QSplitter, QTableView,
    QVBoxLayout, QWidget)
from VeraGrid.Gui.Icons.icons_rc import *

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName(u"MainWindow")
        MainWindow.resize(836, 561)
        icon = QIcon()
        icon.addFile(u":/Icons/icons/solar_power.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        MainWindow.setWindowIcon(icon)
        self.actionCopy_to_clipboard = QAction(MainWindow)
        self.actionCopy_to_clipboard.setObjectName(u"actionCopy_to_clipboard")
        icon1 = QIcon()
        icon1.addFile(u":/Icons/icons/copy.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.actionCopy_to_clipboard.setIcon(icon1)
        self.actionSave = QAction(MainWindow)
        self.actionSave.setObjectName(u"actionSave")
        icon2 = QIcon()
        icon2.addFile(u":/Icons/icons/import_profiles.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.actionSave.setIcon(icon2)
        self.verticalLayout = QVBoxLayout(MainWindow)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.splitter = QSplitter(MainWindow)
        self.splitter.setObjectName(u"splitter")
        self.splitter.setOrientation(Qt.Horizontal)
        self.frame = QFrame(self.splitter)
        self.frame.setObjectName(u"frame")
        self.frame.setFrameShape(QFrame.NoFrame)
        self.frame.setFrameShadow(QFrame.Raised)
        self.gridLayout_2 = QGridLayout(self.frame)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.label_3 = QLabel(self.frame)
        self.label_3.setObjectName(u"label_3")

        self.gridLayout_2.addWidget(self.label_3, 2, 0, 1, 2)

        self.label_2 = QLabel(self.frame)
        self.label_2.setObjectName(u"label_2")

        self.gridLayout_2.addWidget(self.label_2, 3, 0, 1, 1)

        self.label_5 = QLabel(self.frame)
        self.label_5.setObjectName(u"label_5")

        self.gridLayout_2.addWidget(self.label_5, 7, 0, 1, 1)

        self.latitudeSpinBox = QDoubleSpinBox(self.frame)
        self.latitudeSpinBox.setObjectName(u"latitudeSpinBox")
        self.latitudeSpinBox.setDecimals(6)
        self.latitudeSpinBox.setMinimum(-9999999.000000000000000)
        self.latitudeSpinBox.setMaximum(99999.000000000000000)

        self.gridLayout_2.addWidget(self.latitudeSpinBox, 6, 1, 1, 2)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.gridLayout_2.addItem(self.verticalSpacer, 9, 0, 1, 1)

        self.longitudeSpinBox = QDoubleSpinBox(self.frame)
        self.longitudeSpinBox.setObjectName(u"longitudeSpinBox")
        self.longitudeSpinBox.setDecimals(6)
        self.longitudeSpinBox.setMinimum(-9999999.000000000000000)
        self.longitudeSpinBox.setMaximum(99999.000000000000000)

        self.gridLayout_2.addWidget(self.longitudeSpinBox, 7, 1, 1, 2)

        self.powerSpinBox = QDoubleSpinBox(self.frame)
        self.powerSpinBox.setObjectName(u"powerSpinBox")
        self.powerSpinBox.setMaximum(99999.000000000000000)

        self.gridLayout_2.addWidget(self.powerSpinBox, 8, 1, 1, 2)

        self.label_4 = QLabel(self.frame)
        self.label_4.setObjectName(u"label_4")

        self.gridLayout_2.addWidget(self.label_4, 6, 0, 1, 1)

        self.label = QLabel(self.frame)
        self.label.setObjectName(u"label")

        self.gridLayout_2.addWidget(self.label, 8, 0, 1, 1)

        self.label_bus = QLabel(self.frame)
        self.label_bus.setObjectName(u"label_bus")

        self.gridLayout_2.addWidget(self.label_bus, 5, 0, 1, 3)

        self.label_gen = QLabel(self.frame)
        self.label_gen.setObjectName(u"label_gen")

        self.gridLayout_2.addWidget(self.label_gen, 4, 0, 1, 3)

        self.generateButton = QPushButton(self.frame)
        self.generateButton.setObjectName(u"generateButton")

        self.gridLayout_2.addWidget(self.generateButton, 10, 2, 1, 1)

        self.splitter.addWidget(self.frame)
        self.frame_4 = QFrame(self.splitter)
        self.frame_4.setObjectName(u"frame_4")
        self.frame_4.setFrameShape(QFrame.NoFrame)
        self.frame_4.setFrameShadow(QFrame.Raised)
        self.gridLayout = QGridLayout(self.frame_4)
        self.gridLayout.setObjectName(u"gridLayout")
        self.plotButton = QPushButton(self.frame_4)
        self.plotButton.setObjectName(u"plotButton")
        icon3 = QIcon()
        icon3.addFile(u":/Icons/icons/plot.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.plotButton.setIcon(icon3)

        self.gridLayout.addWidget(self.plotButton, 1, 0, 1, 2)

        self.horizontalSpacer_2 = QSpacerItem(394, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.gridLayout.addItem(self.horizontalSpacer_2, 1, 2, 1, 1)

        self.acceptButton = QPushButton(self.frame_4)
        self.acceptButton.setObjectName(u"acceptButton")

        self.gridLayout.addWidget(self.acceptButton, 1, 3, 1, 1)

        self.resultsTableView = QTableView(self.frame_4)
        self.resultsTableView.setObjectName(u"resultsTableView")

        self.gridLayout.addWidget(self.resultsTableView, 0, 0, 1, 4)

        self.splitter.addWidget(self.frame_4)

        self.verticalLayout.addWidget(self.splitter)


        self.retranslateUi(MainWindow)

        QMetaObject.connectSlotsByName(MainWindow)
    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QCoreApplication.translate("MainWindow", u"Solar power wizard", None))
        self.actionCopy_to_clipboard.setText(QCoreApplication.translate("MainWindow", u"Copy to clipboard", None))
        self.actionSave.setText(QCoreApplication.translate("MainWindow", u"Save", None))
        self.label_3.setText(QCoreApplication.translate("MainWindow", u"Site data", None))
        self.label_2.setText("")
        self.label_5.setText(QCoreApplication.translate("MainWindow", u"Longitude", None))
        self.latitudeSpinBox.setSuffix(QCoreApplication.translate("MainWindow", u" deg", None))
        self.longitudeSpinBox.setSuffix(QCoreApplication.translate("MainWindow", u" deg", None))
        self.powerSpinBox.setSuffix(QCoreApplication.translate("MainWindow", u" MW", None))
        self.label_4.setText(QCoreApplication.translate("MainWindow", u"Latutide", None))
        self.label.setText(QCoreApplication.translate("MainWindow", u"Power", None))
        self.label_bus.setText(QCoreApplication.translate("MainWindow", u"Bus", None))
        self.label_gen.setText(QCoreApplication.translate("MainWindow", u"Generator", None))
        self.generateButton.setText(QCoreApplication.translate("MainWindow", u"Generate", None))
#if QT_CONFIG(tooltip)
        self.plotButton.setToolTip(QCoreApplication.translate("MainWindow", u"Plot data", None))
#endif // QT_CONFIG(tooltip)
        self.plotButton.setText("")
        self.acceptButton.setText(QCoreApplication.translate("MainWindow", u"Accept", None))
    # retranslateUi




# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.  
# SPDX-License-Identifier: MPL-2.0


import sys
from typing import List
import pandas as pd
from PySide6 import QtWidgets

from VeraGrid.Gui.TowerBuilder.tower_builder import Ui_TowerBuilderDialog
import VeraGridEngine.Devices as dev
from VeraGrid.Gui.TowerBuilder.table_models import TowerModel, WireInTower, WiresTable, Wire
from VeraGrid.Gui.pandas_model import PandasModel
from VeraGrid.Gui.general_dialogues import LogsDialogue
from VeraGridEngine.basic_structures import Logger


class TowerBuilderGUI(QtWidgets.QDialog):

    def __init__(self, tower: dev.OverheadLineType = None, wires_catalogue: List[Wire] = None):
        """
        Constructor
        :param tower:
        :param wires_catalogue:
        """
        QtWidgets.QDialog.__init__(self)
        self.ui = Ui_TowerBuilderDialog()
        self.ui.setupUi(self)
        self.setWindowTitle('Line builder')

        # create wire collection from the catalogue
        self.wires_table = WiresTable(self)

        # create the tower driver
        self.tower_driver = TowerModel(edit_callback=self.compute, tower=tower)

        self.ui.name_lineEdit.setText(self.tower_driver.tower.name)
        self.ui.rho_doubleSpinBox.setValue(self.tower_driver.tower.earth_resistivity)
        self.ui.frequency_doubleSpinBox.setValue(self.tower_driver.tower.frequency)
        self.ui.voltage_doubleSpinBox.setValue(self.tower_driver.tower.Vnom)

        if wires_catalogue is not None:
            for wire in wires_catalogue:
                self.wires_table.add(wire)
            self.compute()

        # matrix combo
        self.ui.matrixViewComboBox.addItem("Series impedance [Ω/km]")
        self.ui.matrixViewComboBox.addItem("Series impedance (no neutral) [Ω/km]")
        self.ui.matrixViewComboBox.addItem("Series impedance (sequence) [Ω/km]")
        self.ui.matrixViewComboBox.addItem("Shunt admittance [μS/km]")
        self.ui.matrixViewComboBox.addItem("Shunt admittance (no neutral) [μS/km]")
        self.ui.matrixViewComboBox.addItem("Shunt admittance (sequence) [μS/km]")


        # set models
        self.ui.wires_tableView.setModel(self.wires_table)
        self.ui.tower_tableView.setModel(self.tower_driver)

        # set divider
        self.ui.main_splitter.setStretchFactor(0, 6)
        self.ui.main_splitter.setStretchFactor(1, 2)

        # button clicks
        self.ui.add_to_tower_pushButton.clicked.connect(self.add_wire_to_tower)
        self.ui.delete_from_tower_pushButton.clicked.connect(self.delete_wire_from_tower)
        self.ui.compute_pushButton.clicked.connect(self.compute_btn_click)
        self.ui.acceptButton.clicked.connect(self.accept)
        self.ui.name_lineEdit.textChanged.connect(self.name_changed)

        # combobox update
        self.ui.matrixViewComboBox.currentIndexChanged.connect(self.show_matrix)
        self.ui.frequency_doubleSpinBox.valueChanged.connect(self.compute)
        self.ui.rho_doubleSpinBox.valueChanged.connect(self.compute)
        self.ui.voltage_doubleSpinBox.valueChanged.connect(self.compute)

    def msg(self, text, title="Warning"):
        """
        Message box
        :param text: Text to display
        :param title: Name of the window
        """
        msg = QtWidgets.QMessageBox()
        msg.setIcon(QtWidgets.QMessageBox.Icon.Information)
        msg.setText(text)
        # msg.setInformativeText("This is additional information")
        msg.setWindowTitle(title)
        # msg.setDetailedText("The details are as follows:")
        msg.setStandardButtons(QtWidgets.QMessageBox.StandardButton.Ok)
        retval = msg.exec()

    def name_changed(self):
        """
        Change name
        :return:
        """
        self.tower_driver.tower.name = self.ui.name_lineEdit.text()

    def add_wire_to_collection(self):
        """
        Add new wire to collection
        :return:
        """
        name = 'Wire_' + str(len(self.wires_table.wires) + 1)
        wire = dev.Wire(name=name, r_ext=0.01, r=0.01, x=0)
        self.wires_table.add(wire)

    def delete_wire_from_collection(self):
        """
        Delete wire from the collection
        :return:
        """
        idx = self.ui.wires_tableView.currentIndex()
        sel_idx = idx.row()

        if sel_idx > -1:

            # delete_with_dialogue all the wires from the tower too
            self.tower_driver.delete_by_name(self.wires_table.wires[sel_idx])

            # delete_with_dialogue from the catalogue
            self.wires_table.delete(sel_idx)

            self.compute()
        else:
            self.msg('Select a wire in the wires catalogue')

    def add_wire_to_tower(self):
        """
        Add wire to tower
        :return:
        """
        idx = self.ui.wires_tableView.currentIndex()
        sel_idx = idx.row()

        if sel_idx > -1:
            selected_wire: Wire = self.wires_table.wires[sel_idx].copy()
            self.tower_driver.add(WireInTower(wire=selected_wire))
            self.compute()
        else:
            self.msg('Select a wire in the wires catalogue')

    def delete_wire_from_tower(self):
        """
        Delete wire from the tower
        :return:
        """
        idx = self.ui.tower_tableView.currentIndex()
        sel_idx = idx.row()

        if sel_idx > -1:
            self.tower_driver.delete(sel_idx)
            self.compute()
        else:
            self.msg('Select a wire from the wire composition')

    def show_matrix(self):
        """
        Display a computed matrix
        :return:
        """
        idx = self.ui.matrixViewComboBox.currentIndex()

        if idx == 0:
            # Impedances in Ohm/km
            cols = ['Phase' + str(i) for i in self.tower_driver.tower.z_phases_nabc]
            z_df = pd.DataFrame(data=self.tower_driver.tower.z_nabc, columns=cols, index=cols)
            self.ui.matrixTableView.setModel(PandasModel(z_df))

        elif idx == 1:

            cols = ['Phase' + str(i) for i in self.tower_driver.tower.z_phases_abc]
            z_df = pd.DataFrame(data=self.tower_driver.tower.z_abc, columns=cols, index=cols)
            self.ui.matrixTableView.setModel(PandasModel(z_df))

        elif idx == 2:
            if self.tower_driver.tower.z_seq is not None:
                if self.tower_driver.tower.z_seq.shape[0] % 3 == 0:
                    ncirc = max(self.tower_driver.tower.wires_in_tower.get_circuits())
                    cols = [f'Seq{i}@circ{c + 1}' for i in range(3) for c in range(ncirc)]
                    z_df = pd.DataFrame(data=self.tower_driver.tower.z_seq, columns=cols, index=cols)
                    self.ui.matrixTableView.setModel(PandasModel(z_df))
                else:
                    self.ui.matrixTableView.setModel(None)
            else:
                self.ui.matrixTableView.setModel(None)

        elif idx == 3:
            # Admittances in uS/km
            cols = ['Phase' + str(i) for i in self.tower_driver.tower.y_phases_abcn]
            z_df = pd.DataFrame(data=self.tower_driver.tower.y_abcn.imag * 1e6, columns=cols, index=cols)
            self.ui.matrixTableView.setModel(PandasModel(z_df))

        elif idx == 4:
            cols = ['Phase' + str(i) for i in self.tower_driver.tower.y_phases_abc]
            z_df = pd.DataFrame(data=self.tower_driver.tower.y_abc.imag * 1e6, columns=cols, index=cols)
            self.ui.matrixTableView.setModel(PandasModel(z_df))

        elif idx == 5:
            if self.tower_driver.tower.y_seq is not None:
                if self.tower_driver.tower.y_seq.shape[0] % 3 == 0:
                    ncirc = max(self.tower_driver.tower.wires_in_tower.get_circuits())
                    cols = [f'Seq{i}@circ{c + 1}' for i in range(3) for c in range(ncirc)]
                    z_df = pd.DataFrame(data=self.tower_driver.tower.y_seq.imag * 1e6, columns=cols, index=cols)
                    self.ui.matrixTableView.setModel(PandasModel(z_df))
                else:
                    self.ui.matrixTableView.setModel(None)
            else:
                self.ui.matrixTableView.setModel(None)

        # set auto adjust headers
        self.ui.matrixTableView.horizontalHeader().setSectionResizeMode(
            QtWidgets.QHeaderView.ResizeMode.ResizeToContents)
        self.ui.matrixTableView.horizontalHeader().setSectionResizeMode(
            QtWidgets.QHeaderView.ResizeMode.ResizeToContents)
        self.ui.matrixTableView.horizontalHeader().setSectionResizeMode(
            QtWidgets.QHeaderView.ResizeMode.ResizeToContents)

        self.ui.matrixTableView.horizontalHeader().setSectionResizeMode(
            QtWidgets.QHeaderView.ResizeMode.ResizeToContents)
        self.ui.matrixTableView.horizontalHeader().setSectionResizeMode(
            QtWidgets.QHeaderView.ResizeMode.ResizeToContents)
        self.ui.matrixTableView.horizontalHeader().setSectionResizeMode(
            QtWidgets.QHeaderView.ResizeMode.ResizeToContents)

    def compute(self):
        """

        :return:
        """

        self.tower_driver.tower.frequency = self.ui.frequency_doubleSpinBox.value()
        self.tower_driver.tower.earth_resistivity = self.ui.rho_doubleSpinBox.value()
        self.tower_driver.tower.Vnom = self.ui.voltage_doubleSpinBox.value()

        # heck the wires configuration
        logs = Logger()
        all_ok = self.tower_driver.tower.check(logs)

        if all_ok:
            # try:
            # compute the matrices
            self.tower_driver.tower.compute()

            self.show_matrix()

            # plot
            self.plot()

            # except Exception as e:
            #     self.msg(str(e), 'Tower calculation')

        return all_ok, logs

    def plot(self):
        """
        PLot the tower distribution
        """
        self.ui.plotwidget.clear()

        fig = self.ui.plotwidget.get_figure()
        fig.set_facecolor('white')
        ax = self.ui.plotwidget.get_axis()
        self.tower_driver.tower.plot(ax=ax)
        fig.tight_layout()
        self.ui.plotwidget.redraw()

    def compute_btn_click(self):
        """
        Call compute and display the bugs (if any)
        :return:
        """
        all_ok, logs = self.compute()

        if not all_ok:
            logger_diag = LogsDialogue(name='Tower computation', logger=logs)
            logger_diag.exec()

    def example_1(self):
        name = '4/0 6/1 ACSR'
        r = 0.367851632  # ohm / km
        x = 0  # ohm / km
        gmr = 0.002481072  # m

        wire = dev.Wire(name=name, r_ext=gmr, r=r, x=x)

        self.tower_driver.add(WireInTower(wire=wire, xpos=0, ypos=8.8392, phase=0))
        self.tower_driver.add(WireInTower(wire=wire, xpos=0.762, ypos=8.8392, phase=1))
        self.tower_driver.add(WireInTower(wire=wire, xpos=2.1336, ypos=8.8392, phase=2))
        self.tower_driver.add(WireInTower(wire=wire, xpos=1.2192, ypos=7.62, phase=3))

        self.wires_table.add(wire)

    def example_2(self):
        name = '4/0 6/1 ACSR'
        r = 0.367851632  # ohm / km
        x = 0  # ohm / km
        gmr = 0.002481072  # m

        incx = 0.1
        incy = 0.1

        wire = dev.Wire(name=name, r_ext=gmr, r=r, x=x)

        self.tower_driver.add(WireInTower(wire, xpos=0, ypos=8.8392, phase=1))
        self.tower_driver.add(WireInTower(wire, xpos=0.762, ypos=8.8392, phase=2))
        self.tower_driver.add(WireInTower(wire, xpos=2.1336, ypos=8.8392, phase=3))
        self.tower_driver.add(WireInTower(wire, xpos=1.2192, ypos=7.62, phase=0))

        self.tower_driver.add(WireInTower(wire, xpos=incx + 0, ypos=8.8392, phase=1))
        self.tower_driver.add(WireInTower(wire, xpos=incx + 0.762, ypos=8.8392, phase=2))
        self.tower_driver.add(WireInTower(wire, xpos=incx + 2.1336, ypos=8.8392, phase=3))
        # self.tower.add(Wire(name, xpos=incx+1.2192, ypos=7.62, gmr=gmr, r=r, x=x, phase=0))

        self.tower_driver.add(WireInTower(wire, xpos=incx / 2 + 0, ypos=incy + 8.8392, phase=1))
        self.tower_driver.add(WireInTower(wire, xpos=incx / 2 + 0.762, ypos=incy + 8.8392, phase=2))
        self.tower_driver.add(WireInTower(wire, xpos=incx / 2 + 2.1336, ypos=incy + 8.8392, phase=3))
        # self.tower.add(Wire(name, xpos=incx/2 + 1.2192, ypos=incy+7.62, gmr=gmr, r=r, x=x, phase=0))

        self.wires_table.add(wire)

    def example_3(self):
        name = '4/0 6/1 ACSR'
        r = 0.367851632  # ohm / km
        x = 0  # ohm / km
        gmr = 0.002481072  # m

        wire = dev.Wire(name=name, r_ext=gmr, r=r, x=x)

        self.tower_driver.add(WireInTower(wire=wire, xpos=1.5, ypos=10, phase=0))
        self.tower_driver.add(WireInTower(wire=wire, xpos=1, ypos=7, phase=1))
        self.tower_driver.add(WireInTower(wire=wire, xpos=1, ypos=8, phase=2))
        self.tower_driver.add(WireInTower(wire=wire, xpos=1, ypos=9, phase=3))

        self.tower_driver.add(WireInTower(wire=wire, xpos=2, ypos=7, phase=4))
        self.tower_driver.add(WireInTower(wire=wire, xpos=2, ypos=8, phase=5))
        self.tower_driver.add(WireInTower(wire=wire, xpos=2, ypos=9, phase=6))

        self.wires_table.add(wire)


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    window = TowerBuilderGUI()

    window.example_3()
    window.compute()

    window.resize(1.61 * 600.0, 600.0)  # golden ratio
    window.show()
    sys.exit(app.exec())



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
import os
import sys
import pandas as pd
from PySide6 import QtGui, QtCore, QtWidgets
from typing import List, Tuple, Union
from VeraGrid.Gui.CoordinatesInput.coordinates_input_gui import Ui_Dialog
from VeraGrid.Gui.ProfilesInput.excel_dialog import ExcelDialog
from VeraGrid.Gui.general_dialogues import LogsDialogue
from VeraGridEngine.Devices.Substation import Bus
from VeraGridEngine.basic_structures import Logger
from VeraGrid.Gui.messages import error_msg


def get_list_model(iterable: List[str]) -> QtGui.QStandardItemModel:
    """
    get Qt list model from a simple iterable
    :param iterable: 
    :return: List model
    """
    list_model = QtGui.QStandardItemModel()
    if iterable is not None:
        for val in iterable:
            # for the list model
            item = QtGui.QStandardItem(str(val))
            item.setEditable(False)
            list_model.appendRow(item)
    return list_model


def find_duplicates(arr: List[str]) -> Tuple[List[str], List[str]]:
    """
    Find duplicates in an array
    :param arr: original array
    :return: unique, duplicates
    """
    seen = set()
    duplicates = set()

    for element in arr:
        if element in seen:
            duplicates.add(element)
        else:
            seen.add(element)

    return list(seen), list(duplicates)


class CoordinatesInputAssociation:
    """
    Association
    """
    def __init__(self, name, code, x=0, y=0, latitude=0, longitude=0):
        """

        :param name:
        :param code:
        :param x:
        :param y:
        :param latitude:
        :param longitude:
        """
        self.name: str = name
        self.code: str = code
        self.x: float = x
        self.y: float = y
        self.latitude: float = latitude
        self.longitude: float = longitude

    def get_at(self, idx: int) -> Union[float, str]:
        """
        Get association at index
        :param idx:
        :return:
        """
        if idx == 0:
            return self.name
        elif idx == 1:
            return self.code
        elif idx == 2:
            return self.x
        elif idx == 3:
            return self.y
        elif idx == 4:
            return self.latitude
        elif idx == 5:
            return self.longitude
        else:
            return ''


class CoordinatesInputAssociations(QtCore.QAbstractTableModel):

    def __init__(self):
        QtCore.QAbstractTableModel.__init__(self)

        self.__values: List[CoordinatesInputAssociation] = list()

        self.__headers = ['Name', 'Code', 'x', 'y', 'latitude', 'longitude']

    def update(self):
        """
        update table
        """
        self.layoutAboutToBeChanged.emit()
        self.layoutChanged.emit()

    def append(self, val: CoordinatesInputAssociation):
        self.__values.append(val)
        self.update()

    def set_x_at(self, idx, value):
        self.__values[idx].x = value
        self.update()

    def set_y_at(self, idx, value):
        self.__values[idx].y = value
        self.update()

    def set_latitude_at(self, idx, value):
        self.__values[idx].latitude = value
        self.update()

    def set_longitude_at(self, idx, value):
        self.__values[idx].longitude = value
        self.update()

    def get_name_at(self, idx):
        return self.__values[idx].name

    def get_code_at(self, idx):
        return self.__values[idx].code

    def get_x_at(self, idx):
        return self.__values[idx].x

    def get_y_at(self, idx):
        return self.__values[idx].y

    def get_latitude_at(self, idx):
        return self.__values[idx].latitude

    def get_longitude_at(self, idx):
        return self.__values[idx].longitude

    def clear_at(self, idx):
        self.__values[idx].x = 0
        self.__values[idx].y = 0
        self.__values[idx].latitude = 0
        self.__values[idx].longitude = 0
        self.update()

    def rowCount(self, parent=None):
        return len(self.__values)

    def columnCount(self, parent=None):
        return len(self.__headers)

    def data(self, index, role=QtCore.Qt.ItemDataRole.DisplayRole):
        if index.isValid():
            if role == QtCore.Qt.ItemDataRole.DisplayRole:
                return str(self.__values[index.row()].get_at(index.column()))
        return None

    def headerData(self,
                   section: int,
                   orientation: QtCore.Qt.Orientation,
                   role=QtCore.Qt.ItemDataRole.DisplayRole):

        if role == QtCore.Qt.ItemDataRole.DisplayRole:
            if orientation == QtCore.Qt.Orientation.Horizontal:
                return self.__headers[section]
            elif orientation == QtCore.Qt.Orientation.Vertical:
                return section
        return None


class CoordinatesInputGUI(QtWidgets.QDialog):
    """
    CoordinatesInputGUI
    """

    def __init__(self, parent=None, list_of_objects: List[Bus] = list()):
        """

        Args:
            parent:
            list_of_objects: List of objects to which set a profile to
            list_of_objects: list ob object to modify
        """
        QtWidgets.QDialog.__init__(self, parent)
        if list_of_objects is None:
            list_of_objects = list()
        self.ui = Ui_Dialog()
        self.ui.setupUi(self)
        self.setWindowTitle('Coordinates import dialogue')
        self.setAcceptDrops(True)

        self.project_directory = None

        self.assigned_count: int = 0

        # dataFrame
        self.original_data_frame: pd.DataFrame | None = None

        # initialize the objectives list
        self.objects: List[Bus] = list_of_objects

        self.associations = CoordinatesInputAssociations()
        for elm in list_of_objects:
            self.associations.append(CoordinatesInputAssociation(elm.name, elm.code, elm.x, elm.y, elm.latitude, elm.longitude))

        self.display_associations()

        self.accepted_extensions = ['.csv', '.xlsx', '.xls']

        self.ui.splitter.setStretchFactor(0, 3)
        self.ui.splitter.setStretchFactor(1, 7)

        # click
        self.ui.open_button.clicked.connect(self.import_profile)
        self.ui.acceptButton.clicked.connect(self.do_it)
        self.ui.refreshButton.clicked.connect(self.assign)

    def msg(self, text, title="Warning"):
        """
        Message box
        :param text: Text to display
        :param title: Name of the window
        """
        msg = QtWidgets.QMessageBox()
        msg.setIcon(QtWidgets.QMessageBox.Icon.Information)
        msg.setText(text)
        # msg.setInformativeText("This is additional information")
        msg.setWindowTitle(title)
        # msg.setDetailedText("The details are as follows:")
        msg.setStandardButtons(QtWidgets.QMessageBox.StandardButton.Ok)
        retval = msg.exec()

    def dropEvent(self, event):
        """
        Drop file on the GUI, the default behaviour is to load the file
        :param event: event containing all the information
        """
        if event.mimeData().hasUrls:
            events = event.mimeData().urls()
            if len(events) > 0:

                file_names = list()

                for event in events:
                    file_name = event.toLocalFile()
                    name, file_extension = os.path.splitext(file_name)
                    if file_extension.lower() in self.accepted_extensions:
                        file_names.append(file_name)
                    else:
                        error_msg('The file type ' + file_extension.lower() + ' is not accepted :(')

                if len(file_names) > 0:
                    # Just open the file
                    self.open_file_now(filename=file_names)

    def display_associations(self):

        self.ui.assignation_table.setModel(self.associations)

    def set_combo_boxes(self):
        """
        Set the options
        """
        if self.original_data_frame is not None:
            names = [val.lower().strip() for val in self.original_data_frame.columns.values]

            mdl = get_list_model(names)
            self.ui.nameComboBox.setModel(mdl)
            self.ui.codeComboBox.setModel(mdl)
            self.ui.xComboBox.setModel(mdl)
            self.ui.yComboBox.setModel(mdl)
            self.ui.latitudeComboBox.setModel(mdl)
            self.ui.longitudeComboBox.setModel(mdl)

            for i, name in enumerate(names):  # names are already in lowercase

                if name in ['x', 'xpos', 'x_pos']:
                    self.ui.xComboBox.setCurrentIndex(i)
                    self.ui.xCheckBox.setChecked(True)

                elif name in ['y', 'ypos', 'y_pos']:
                    self.ui.yComboBox.setCurrentIndex(i)
                    self.ui.yCheckBox.setChecked(True)

                elif name in ['latitude', 'lat']:
                    self.ui.latitudeComboBox.setCurrentIndex(i)
                    self.ui.latitudeCheckBox.setChecked(True)

                elif name in ['longitude', 'lon']:
                    self.ui.longitudeComboBox.setCurrentIndex(i)
                    self.ui.longitudeCheckBox.setChecked(True)

    def assign(self):
        """
        Assign the loaded values to the correspondance table
        """
        if self.original_data_frame is not None:
            name_idx = self.ui.nameComboBox.currentIndex()
            code_idx = self.ui.codeComboBox.currentIndex()
            x_idx = self.ui.xComboBox.currentIndex()
            y_idx = self.ui.yComboBox.currentIndex()
            lat_idx = self.ui.latitudeComboBox.currentIndex()
            lon_idx = self.ui.longitudeComboBox.currentIndex()

            for i in range(self.original_data_frame.shape[0]):

                if self.ui.nameRadioButton.isChecked():
                    if isinstance(self.original_data_frame.values[i, name_idx], float):
                        dat_name = str(int(self.original_data_frame.values[i, name_idx]))
                    else:
                        dat_name = str(self.original_data_frame.values[i, name_idx])

                elif self.ui.codeRadioButton.isChecked():
                    if isinstance(self.original_data_frame.values[i, code_idx], float):
                        dat_name = str(int(self.original_data_frame.values[i, code_idx]))
                    else:
                        dat_name = str(self.original_data_frame.values[i, code_idx])

                else:
                    dat_name = 'a'

                # search in the buses
                for j in range(self.associations.rowCount()):

                    if self.ui.nameRadioButton.isChecked():
                        name = str(self.associations.get_name_at(j))
                    elif self.ui.codeRadioButton.isChecked():
                        name = str(self.associations.get_code_at(j))
                    else:
                        name = 'b'

                    # if the search criteria match...
                    if dat_name.lower() == name.lower():

                        if self.ui.xCheckBox.isChecked():
                            self.associations.set_x_at(j, self.original_data_frame.values[i, x_idx])

                        if self.ui.yCheckBox.isChecked():
                            self.associations.set_y_at(j, self.original_data_frame.values[i, y_idx])

                        if self.ui.longitudeCheckBox.isChecked():
                            self.associations.set_longitude_at(j, self.original_data_frame.values[i, lon_idx])

                        if self.ui.latitudeCheckBox.isChecked():
                            self.associations.set_latitude_at(j, self.original_data_frame.values[i, lat_idx])

            self.display_associations()

            self.assigned_count += 1

    def import_profile(self):
        """
        Select a file to be loaded
        """

        # declare the allowed file types
        files_types = "Formats (*.xlsx *.xls *.csv)"
        # call dialog to select the file
        # filename, type_selected = QFileDialog.getOpenFileNameAndFilter(self, 'Save file', '', files_types)

        # call dialog to select the file
        filename, type_selected = QtWidgets.QFileDialog.getOpenFileName(self, 'Open file', filter=files_types)
        self.open_file_now(filename)

    def open_file_now(self, filename: str):
        """
        Opena file
        :param filename: path of the file
        """
        if len(filename) > 0:
            # get the filename extension
            name, file_extension = os.path.splitext(filename)

            # Depending on the extension load the file
            if file_extension == '.csv':
                self.original_data_frame = pd.read_csv(filename, index_col=None)

            elif file_extension in ['.xlsx', '.xls']:

                # select the sheet from the file
                excel_window = ExcelDialog(self, filename)
                excel_window.exec()
                sheet_index = excel_window.excel_sheet

                if sheet_index is not None:
                    self.original_data_frame = pd.read_excel(filename, sheet_name=sheet_index, index_col=None)
                else:
                    return

            # check for duplicates
            unique_hdr, duplicate_hdr = find_duplicates(arr=list(self.original_data_frame.columns))

            if len(duplicate_hdr):
                # notify
                logger = Logger()
                for hdr in duplicate_hdr:
                    logger.add_error("Duplicated header", device=hdr)

                logs_dialogue = LogsDialogue(name="Duplicated headers", logger=logger, expand_all=True)
                logs_dialogue.exec()

                # filter the headers
                self.original_data_frame = self.original_data_frame[unique_hdr]

            # set the profile names list
            self.set_combo_boxes()
            self.assign()
            self.display_associations()
            self.assigned_count = 0

    def do_it(self) -> None:
        """
        Close. The data has to be queried later to the object by the parent by calling get_association_data
        """

        if self.assigned_count == 0:
            # maybe the user clicked on accept without assigning, so assign as a last resort
            self.assign()

        # assign the values back
        for i, bus in enumerate(self.objects):
            if self.ui.xCheckBox.isChecked():
                bus.x = self.associations.get_x_at(i)

            if self.ui.yCheckBox.isChecked():
                bus.y = - self.associations.get_y_at(i)

            if self.ui.longitudeCheckBox.isChecked():
                bus.longitude = self.associations.get_longitude_at(i)

                if bus.substation is not None:
                    if bus.substation.longitude == 0.0:
                        bus.substation.longitude = bus.longitude

            if self.ui.latitudeCheckBox.isChecked():
                bus.latitude = self.associations.get_latitude_at(i)

                if bus.substation is not None:
                    if bus.substation.latitude == 0.0:
                        bus.substation.latitude = bus.latitude

        self.close()


# if __name__ == "__main__":
#     app = QtWidgets.QApplication(sys.argv)
#
#     class TestObj:
#         def __init__(self, name, code, x=0, y=0, latitude=0, longitude=0):
#             self.name = name
#             self.code = code
#             self.x = x
#             self.y = y
#             self.latitude = latitude
#             self.longitude = longitude
#
#
#     window = CoordinatesInputGUI(list_of_objects=[TestObj('Test object', 'code')] * 10)
#     window.resize(1.61 * 700.0, 600.0)  # golden ratio
#     window.show()
#     sys.exit(app.exec())



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from PySide6.QtWidgets import QPushButton, QVBoxLayout, QDialog, QLabel, QComboBox

from VeraGrid.Gui.gui_functions import create_spinbox, create_int_spinbox, get_list_model
from VeraGridEngine.Devices.Branches.tap_changer import TapChanger, TapChangerTypes


class TransformerTapsEditor(QDialog):
    """
    TransformerTapsEditor
    """

    def __init__(self, api_object: TapChanger):
        QDialog.__init__(self)

        self.setWindowTitle("Tap changer editor")

        self.api_object: TapChanger = api_object

        self.layout = QVBoxLayout(self)

        self.asymmetry_angle_spinner = create_spinbox(value=self.api_object.asymmetry_angle,
                                                      minimum=0, maximum=999999, decimals=6)

        self.total_positions_spinner = create_int_spinbox(value=self.api_object.total_positions,
                                                          minimum=0, maximum=999999)

        self.dV_spinner = create_spinbox(value=self.api_object.dV,
                                         minimum=0, maximum=999999, decimals=6)

        self.neutral_position_spinner = create_int_spinbox(value=self.api_object.neutral_position,
                                                           minimum=0, maximum=999999)

        self.tap_position_spinner = create_int_spinbox(value=self.api_object.tap_position,
                                                       minimum=0, maximum=999999)

        self.tap_changer_types = QComboBox()
        self.tap_changer_types_dict = {
            TapChangerTypes.NoRegulation.value: TapChangerTypes.NoRegulation,
            TapChangerTypes.Symmetrical.value: TapChangerTypes.Symmetrical,
            TapChangerTypes.Asymmetrical.value: TapChangerTypes.Asymmetrical,
            TapChangerTypes.VoltageRegulation.value: TapChangerTypes.VoltageRegulation,
        }
        lst = list(self.tap_changer_types_dict.keys())
        self.tap_changer_types.setModel(get_list_model(lst))
        self.tap_changer_types.setCurrentIndex(lst.index(self.api_object.tc_type.value))

        self.layout.addWidget(QLabel("Tap changer type"))
        self.layout.addWidget(self.tap_changer_types)

        self.layout.addWidget(QLabel("Asymetry angle (deg)"))
        self.layout.addWidget(self.asymmetry_angle_spinner)

        self.layout.addWidget(QLabel("Total positions"))
        self.layout.addWidget(self.total_positions_spinner)

        self.layout.addWidget(QLabel("Neutral position"))
        self.layout.addWidget(self.neutral_position_spinner)

        self.layout.addWidget(QLabel("Tap position"))
        self.layout.addWidget(self.tap_position_spinner)

        self.layout.addWidget(QLabel("Voltage increment per position"))
        self.layout.addWidget(self.dV_spinner)

        # accept button
        self.accept_btn = QPushButton()
        self.accept_btn.setText('Accept')
        self.accept_btn.clicked.connect(self.accept_click)
        self.layout.addWidget(self.accept_btn)

    def accept_click(self):
        """
        Accept
        """

        self.api_object.asymmetry_angle = self.asymmetry_angle_spinner.value()
        self.api_object.total_positions = self.total_positions_spinner.value()
        self.api_object.tap_position = self.tap_position_spinner.value()
        self.api_object.dV = self.dV_spinner.value()
        self.api_object.neutral_position = self.neutral_position_spinner.value()

        self.api_object.tc_type = self.tap_changer_types_dict[self.tap_changer_types.currentText()]

        self.api_object.recalc()

        self.close()
